import json
import os
import sys
import writefile as wf
from collections import namedtuple

"""
This File takes the json structure and parses it, to pass it to writefile.py
    Functions: measurementJsonDecod, parameterJsonDecod, functionJsonDecod, dimensionchange, getdatatypes, changerange, 
                getunit, getobjnames, getMeasurements, getParameters, getFunctions, searchlists, getdefaultvalue, 
                iskeythere, getInterfaceParam, getComponentVarList buildComponents,
                instantiate, instantiateComponents, main
                
    Classes: Measurement, Parameter, Function
"""


def measurementJsonDecod(measurementdict):
    """
    Method that takes a measurementdict and creates a construnctor with corresponding parameters.
    :param measurementdict: Measurement variable as Dictionary
    :return: Tuple with all valuessorted according to an object hook
    """
    return namedtuple('Measurement', measurementdict.keys())(*measurementdict.values())


#creates the constructor for the parameter-object with corresponding parameters
def parameterJsonDecod(parameterdict):
    """
      Method that takes a parameterdict and creates a construnctor with corresponding parameters.
    :param parameterdict: Parameter variable as Dictionary
    :return: Tuple with all valuessorted according to an object hook
    """
    return namedtuple('Parameter', parameterdict.keys())(*parameterdict.values())


#creates the constructor for the function-object with corresponding parameters
def functionJsonDecod(functiondict):
    """
    Method that takes a functiondict and creates a construnctor with corresponding parameters.
    :param functiondict: Function as Dictionary
    :return: Tuple with all valuessorted according to an object hook
    """
    return namedtuple('Function', functiondict.keys())(*functiondict.values())


class Measurement:
    """
    Class for measurement variables, is instantiated using MeasurementJsonDecod.
    """
    def __init__(self, name, description, datatype, value, uuid, dimension, unit=None, range=None):
        """
        Initializes Measurement object.
        :param name: name of the measurement in the component
        :param description: description of the measurement
        :param datatype: datatype of the measurement
        :param value: the value of the measurement object which is generated in getdefaultvalue
        :param uuid: the uuid of the measurement
        :param dimension: the dimension of the measurement changed in dimensionchange
        :param unit: the unit of the measurement
        :param range: the range of the measurement changed in changerange
        """
        self.name = name
        self.description = description
        self.datatype = datatype
        self.range = range
        self.uuid = uuid
        self.dimension = dimension
        self.unit = unit
        if self.datatype == "float":
            self.value = float(value)
        else:
            self.value = value


class Parameter:
    """
    Class for Parameter objects, is inatatntiated using parameterJsonDecod.
    """
    def __init__(self, name, description, datatype, dimension, value, uuid, rangevar=None, unit=None, default=None, constant=False):
        """
        Initializes parameter object.
        :param name: parameter name
        :param description: parameter description
        :param datatype: parameter datatype
        :param dimension: parameter dimension changed in dimensionchange
        :param value: parameter value generated in getdefaultvalue
        :param uuid: parameter uuid
        :param rangevar: parameter range (if not known before added in changerange)
        :param unit: parameter unit (if not known before added in changerange)
        :param default: parameter default value
        :param constant: is the parameter constant or not
        """
        self.name = name
        self.description = description
        self.datatype = datatype
        self.range = rangevar
        self.uuid = uuid
        self.dimension = dimension
        self.unit = unit
        self.default = default
        self.constant = constant
        self.value = value


class Function:
    """
    Class for Function objects, is instantiated using JsonDecode functions
    """
    def __init__(self, name, description, uuid, inargsdatatypes = None, outargsdatatypes = None):
        """
        Initializes function object.
        :param name: function name
        :param description: function description
        :param uuid: function uuid
        :param inargsdatatypes: List of Datatypes of all arguments
        :param outargsdatatypes: List of dataytypes of all returns
        """
        self.name = name
        self.description = description
        self.uuid = uuid
        self.inargsdatatypes = inargsdatatypes
        self.outargsdatatypes = outargsdatatypes


def dimensionchange(string):
    """
    Changes the Format of the dimension from [dimension] to dimension
    :param string: dimension of the json object
    :return: dimension without braces or 0 if dimension is an empty list
    """
    dimarr = string["dimension"]
    if type(dimarr) == list:
        if dimarr:
            value = dimarr[0]
        else:
            value = 0
        newdimension = {"dimension": value}
        string.update(newdimension)
    return string


def getdatatypes(argarray):
    """
    Retrieves the datatypes of an arguments of an argumentarry.
    :param argarray: array of argument names
    :return: array of argument datatypes
    """
    returnarr = []
    if len(argarray) > 0:
        for i in argarray:
            returnarr.append(i["datatype"])
        return returnarr
    else:
        return None


def changerange(obj):
    """
    Adds and changes the range attribute to [None, None] for enums and parameters without range.
    :param obj: The Json object that will be changed
    :return: Changed object
    """
    if iskeythere(obj, "range"):
        if obj["datatype"] == "enum":
            obj.update({"range": [None, None]})
    else:
        obj["range"] = [None, None]
    return obj


def getunit(string):
    """
    Adds unit to parameters without unit.
    :param string: The json object that will be changed
    :return: changed object
    """
    if not iskeythere(string, "unit"):
        string["unit"] = "UNITLESS"
    return string


#gets the objectnames for arguments of the function
def getobjnames(argarray):
    """
    Changes the object names of function in order for them to fit the coding standard.
    :param argarray: array of arguments
    :return: array of changed arguments
    """
    returnarr = []
    if len(argarray) > 0:
        for i in argarray:
            name = (i["name"].lower()).replace(" ", "")
            returnarr.append(name)
        return returnarr
    else:
        return None


def getParameters(parametersDict):
    """
    Parses parameterdicts into parameterobjects.
    :param parametersDict: List of parameterdicts
    :return: List of parameter objects
    """
    # extracts the measurements of a Json-Mode
    parameterobj = []
    if len(parametersDict) > 0:
        for i in parametersDict:
            obj = i
            obj = dimensionchange(obj)
            obj["value"] = getdefaultvalue(obj)
            obj = changerange(obj)
            obj = getunit(obj)
            obj = json.dumps(obj)
            parameterobji = json.loads(obj, object_hook=parameterJsonDecod)
            parameterobj.append(parameterobji)
    return parameterobj


def getMeasurements(measurementsDict):
    """
    Parses measurementdicts into measurement objects
    :param measurementsDict: List of measurementdicts
    :return:List of measurement objects
    """
    measurementobj = []
    if len(measurementsDict) > 0:
        for i in measurementsDict:
            obj = i
            obj = dimensionchange(obj)
            obj["value"] = getdefaultvalue(obj)
            obj = changerange(obj)
            if not iskeythere(obj, "unit"):
                obj["unit"] = "UNITLESS"
            obj = json.dumps(obj)
            measurementobji = json.loads(obj, object_hook=measurementJsonDecod)
            measurementobj.append(measurementobji)
    return measurementobj


#returns a list of all the functions as objects
def getFunctions(functionsDict, jsonmeasurements, jsonparameters):
    """
    Parses functiondicts into function objects.
    :param functionsDict: list of functiondicts
    :param jsonmeasurements: List of all measurement elements
    :param jsonparameters: List of all parameter elements
    :return: List of function objects
    """
    functionobj = []
    if len(functionsDict) > 0:
        for i in functionsDict:
            obj = i
            arguments = searchlists(obj["arguments"], jsonmeasurements)
            arguments = arguments + searchlists(obj["arguments"], jsonparameters)
            returns = []
            if len(obj["returns"]) > 0:
                returns = searchlists(obj["returns"], jsonmeasurements) + searchlists(obj["returns"], jsonparameters)
            obj["inargsnames"] = getobjnames(arguments)
            obj["outargsnames"] = getobjnames(returns)
            obj["inargsdatatypes"] = getdatatypes(arguments)
            obj["outargsdatatypes"] = getdatatypes(returns)
            obj = json.dumps(obj)
            functionobji = json.loads(obj, object_hook=functionJsonDecod)
            functionobj.append(functionobji)
    return functionobj


def searchlists(list1, list2):
    """
    Matches the objects of two lists
    :param list1: List of Elements from a Component Format {name: value}
    :param list2: List of all elements of that type in general
    :return: list of Elements from a component in the {elementtype: ,name: , uuid: , ...} format
    """
    objmeasurementlist = []
    if len(list1) <= 0:
        return []
    for input in list1:
        for searchobj in list2:
            if input["value"] == searchobj["uuid"]:
                objmeasurementlist.append(searchobj)
    return objmeasurementlist


def getdefaultvalue(obj):
    """
    Adds a value to a dict. In case of Parameters with a default value it just takes that
    :param obj: the parameterdict before the changes
    :return: the new value
    """
    if iskeythere(obj, "default"):
        if obj["datatype"] == "string" or obj["datatype"] == "enum":
            return "\"" + obj["default"] + "\""
        return obj["default"]
    datatype = obj["datatype"]
    dimension = obj["dimension"]
    if datatype == "float" or datatype == "double" or datatype == "int":
        if dimension > 1:
            value = []
            for i in range(dimension):
                value.append(0)
            return value
        else:
            return 0
    elif datatype == "boolean":
        return True
    elif datatype == "time":
        return "ua.DateTime.now()"
    else:
        return "\"\""


def iskeythere(list1, key):
    """
    Verifies if the key exists in the dict.
    :param list1: list of  all keys from the object
    :param key: searched key
    :return: Boolean value determining if the key is existent in the dict.
    """
    for keysearch in list1:
        if keysearch == key:
            return True
    return False


def getInterfaceparam(list1, interface, searched, jsoncomponents, searchinglist):
    """
    Gets a list of searched elements from a component and their baseComponents
        and  in the Format {elementtype: , name: , uuid: ,...}
    :param list1: list in which the retrieved elements are saved
    :param interface: the component the values belong to
    :param searched: the elementtype that is looked for
    :param jsoncomponents: a list of  all components
    :param searchinglist: a list of all elements of the searched elementtype
    :return: returns the list in which the retrieved elements were saved
    """
    for inputsearch in interface[searched]:
        for searchobj in searchinglist:
            if inputsearch["value"] == searchobj["uuid"]:
                list1.append(searchobj)
    if iskeythere(list(interface.keys()), "baseComponent"):
        for component in jsoncomponents:
            if component["uuid"] == interface["baseComponent"]:
                return getInterfaceparam(list1, component, searched, jsoncomponents, searchinglist)
    else:
        return list1


def getComponentVarList(searched, list1, component, jsoncomponents):
    """
    Seatches recursivly through the component and its baseComponents for elements of the serached type.
    :param searched: Elementtype that is searched
    :param list1: the list in which all retrieved dicts get stored
    :param component: The component which is viewed at the moment
    :param jsoncomponents: a list of all components
    :return:The list with all stored elements. It has the Format {name : value}
    """
    variablelist = component[searched]
    list1 = list1 + variablelist
    if iskeythere(list(component.keys()), "baseComponent"):
        for search in jsoncomponents:
            if search["uuid"] == component["baseComponent"]:
                return getComponentVarList(searched, list1, search, jsoncomponents)
    else:
        return list1


def buildComponents(interfacesreturn,  jsoninterfaces, jsoncomponents, jsonmeasurements, jsonparameters, jsonfunctions,  enumnodedict, enumvalues):
    """objcomponentsclone,
    Iterates recursively throught the rootcomponent of the interface and its components etc.
    With every iteration it collects all variables and functions from each component and their basecomponents,
    creates their respective objects and writesthem with the writefunctions from wf.
    :param interfacesreturn: a dict in which all created objecttype node names get collect in the Format {uuid : objecttype}
    :param jsoninterfaces: a list of all components that get looked at during the one rekursive iteration
    :param jsoncomponents: a list of all elements with type component
    :param jsonmeasurements: a list of all ements with type measurement
    :param jsonparameters: a list with all elements of type parameter
    :param jsonfunctions: a list of all elements with type function
    :param enumnodedict: a dict with all the enumtypes from the format {uuid : enumtypenodename}
    :param enumvalues: a dict with the changed values from enums from the format {uuid: [0: value1, 1: value2....],...}
    :return: Interfacesreturn
    """
    for obj in jsoninterfaces:
        rootobj = {}
        if obj["elementType"] == "interface":
            for search in jsoncomponents:
                if obj["rootComponent"] == search["uuid"]:
                    rootobj = search
        else:
            rootobj = obj
        if obj["name"] in list(interfacesreturn.keys()):
            continue
        objcomponentslist = getInterfaceparam([], rootobj, "components", jsoncomponents, jsoncomponents)
        if obj["name"] not in interfacesreturn.keys():
            interfacesreturn[str(obj["uuid"])] = wf.openobject(obj["name"], rootobj["description"])
            objmeasurementlist = getInterfaceparam([], rootobj, "measurements", jsoncomponents, jsonmeasurements)
            measurementlist = getComponentVarList("measurements", [], rootobj, jsoncomponents)
            parameterlist = getComponentVarList("parameters", [], rootobj, jsoncomponents)
            measurementsobjs = getMeasurements(objmeasurementlist)
            wf.writemeasurement(measurementsobjs, interfacesreturn[str(obj["uuid"])], enumnodedict, enumvalues,  measurementlist.copy())
            objparameterlist = getInterfaceparam([], rootobj, "parameters", jsoncomponents, jsonparameters)
            parameterobjs = getParameters(objparameterlist)
            wf.writeparameter(parameterobjs, interfacesreturn[str(obj["uuid"])], enumnodedict, enumvalues,  parameterlist.copy())
            objfunctionlist = getInterfaceparam([], rootobj, "functions", jsoncomponents, jsonfunctions)
            functionobj = getFunctions(objfunctionlist, jsonparameters, jsonmeasurements)
            functionvars = getComponentVarList("functions", [], rootobj, jsoncomponents)
            wf.writefunction(functionobj, interfacesreturn[str(obj["uuid"])], functionvars.copy())
            buildComponents(interfacesreturn,  objcomponentslist, jsoncomponents, jsonmeasurements, jsonparameters, jsonfunctions,  enumnodedict, enumvalues)
    return interfacesreturn


def instantiateComponents(typesdict, interfaces, jsoncomponents, instantiatedict, eventsdict, streamsdict, componentlist, headnodeindex, variabledict):
    """
    Instantiates all components and collects events, streams and variable names on the side.
    :param typesdict: Dict for objecttype node names from the format {uuid : objecttypename}
    :param interfaces: A list of components
    :param jsoncomponents: all elements of the type component
    :param instantiatedict: A dict for the node names of the already instantiated objects
                                from Format {objectname : object node name}
    :param eventsdict: Dict where the events of each instantiated component get collected
                            from Format {objectname : eventlist}
    :param streamsdict: Dict where the streams of each instantiated component get collected
                            from Format {objectname : streamlist}
    :param componentlist: a list for all component objects of the parentnaode to retrieve the individual object name
    :param headnodeindex: the individual name of the headnode
    :param variabledict: A retrieved dict for all variables of each component
                            from Format {objectname : variablelist}
    :return: instantiateddict, eventsdict, streamsdict, variabledict
    """
    for interface in interfaces:
        root = interface
        interfacename = ""
        newcomponentlist = getComponentVarList("components", [], root, jsoncomponents)
        for cobj in componentlist:
            if cobj["value"] == root["uuid"]:
                interfacename = cobj["name"]
                componentlist.remove(cobj)
                break
        events = getComponentVarList("events", [], root, jsoncomponents)
        streams = getComponentVarList("streams", [], root, jsoncomponents)
        variables = getComponentVarList("measurements", [], root, jsoncomponents) + getComponentVarList("parameters", [], root, jsoncomponents)
        variabledict[interfacename] = variables
        components = getInterfaceparam([], root, "components", jsoncomponents, jsoncomponents)
        instantiatedict[str(interfacename)] = wf.writeobjects(typesdict[str(interface["uuid"])], interfacename, instantiatedict[str(headnodeindex)])
        eventsdict[str(interfacename)] = events
        streamsdict[str(interfacename)] = streams
        if len(components) > 0:
            instantiateComponents(typesdict, components, jsoncomponents, instantiatedict, eventsdict, streamsdict, newcomponentlist, interfacename, variabledict)

    return instantiatedict, eventsdict, streamsdict, variabledict


def instantiate(typesdict, interfaces, jsoncomponents):
    """
    Instantiates the interface root and starts instantiateComponents
    :param typesdict: Dict for objecttype node names from the format {uuid : objecttypename}.
    :param interfaces: A list of interfaces
    :param jsoncomponents: A list of all componentdicts
    :return: the values from instantiateComponents
    """
    instantiatedict = {}
    eventsdict = {}
    streamsdict ={}
    variabledict = {}
    for interface in interfaces:
        root = {}
        for component in jsoncomponents:
            if interface["rootComponent"] == component["uuid"]:
                root = component
        events = getComponentVarList("events", [], root, jsoncomponents)
        streams = getComponentVarList("streams", [], root, jsoncomponents)
        variables = getComponentVarList("measurements", [], root, jsoncomponents) + getComponentVarList("parameters", [], root, jsoncomponents)
        variabledict[interface["uuid"]] = variables
        components = getInterfaceparam([], root, "components", jsoncomponents, jsoncomponents)
        instantiatedict[str(interface["uuid"])] = wf.writeinterface(typesdict[str(interface["uuid"])], interface["name"])
        eventsdict[str(interface["uuid"])] = events
        streamsdict[str(interface["uuid"])] = streams
        newcomponentlist = getComponentVarList("components", [], root, jsoncomponents)
        return instantiateComponents(typesdict, components, jsoncomponents, instantiatedict, eventsdict, streamsdict, newcomponentlist, interface["uuid"], variabledict)


def main():
    """
    Main method that collects everything and starts all writing process.
    :return: None
    """
    try:
        JSON_MODEL = sys.argv[1][:-5]
        if not os.path.exists(os.path.join('..', 'soil-opc-ua', f'{JSON_MODEL}.json')):
            raise Exception()
    except Exception as e:
        print('You must provide a file from the folder to be used as root file of the parsing, e.g., "python jsonparser.py model.json".')
        exit()

    with open(f'{JSON_MODEL}.json') as file:

        wf.writebeginning()
        jsonobj = json.loads(file.read())
        elements = jsonobj["elements"]
        if len(elements) <= 0:
            exit()
        jsonmeasurements = []
        jsonparameters = []
        jsonfunctions = []
        jsoncomponents = []
        jsoninterfaces = []
        jsonenums = []
        for obj in elements:
            if obj["elementType"] == "measurement":
                jsonmeasurements.append(obj)
            elif obj["elementType"] == "parameter":
                jsonparameters.append(obj)
            elif obj["elementType"] == "function":
                jsonfunctions.append(obj)
            elif obj["elementType"] == "component":
                jsoncomponents.append(obj)
            elif obj["elementType"] == "interface":
                jsoninterfaces.append(obj)
            elif obj["elementType"] == "enum":
                jsonenums.append(obj)
        returnvalenums = wf.writeenums(jsonenums)
        enumnodedict = returnvalenums[0]
        enumvalues = returnvalenums[1]
        typesdict = buildComponents({}, jsoninterfaces, jsoncomponents, jsonmeasurements, jsonparameters, jsonfunctions, enumnodedict, enumvalues)
        returnval = instantiate(typesdict, jsoninterfaces, jsoncomponents)
        instantiatedict = returnval[0]
        eventsdict = returnval[1]
        streamsdict = returnval[2]
        variabledict = returnval[3]
        generatordict = {}
        streamgendict = {}
        savestreamvaldict = {}
        for key in list(eventsdict.keys()):
            eventlist = wf.writeevents(eventsdict[key], instantiatedict[key], {})
            generatordict.update(eventlist)

        for key in list(streamsdict.keys()):
            returnval = wf.writestreams(streamsdict[key], instantiatedict[key], {}, variabledict[key], jsonmeasurements, jsonparameters)
            streamgendict.update(returnval[0])
            savestreamvaldict.update(returnval[1])

        wf.writendpt1()

        for key in list(eventsdict.keys()):
            wf.triggerevents(generatordict, eventsdict[key], instantiatedict[key], variabledict[key])

        for key in list(streamsdict.keys()):
            wf.triggerstreams(streamgendict, streamsdict[key], instantiatedict[key], variabledict[key], savestreamvaldict)
        wf.writendpt2()


if __name__ == '__main__':
    main()
